import os
import re
import inspect
from pydantic import BaseModel
from importlib import import_module
from typing import Dict, Literal, Optional, List
from jinja2 import Environment, FileSystemLoader
from .util import normalize_model_name
from . import get_templates_dir


FD_EXPR = re.compile(r"^(\w+)(:(\w+)(\[([^\]]+)\])?)?(!?)$")
SUBTYPED_TYPES = {"ListField", "DictField", "ReferenceField"}
FieldType = Literal[
    "StringField",
    "IntField",
    "FloatField",
    "BoolField",
    "ListField",
    "DictField",
    "ReferenceField",
    "SelfReferenceField",
    "ObjectIdField",
    "DatetimeField",
]


class FieldDescriptor(BaseModel):
    name: str
    field_type: FieldType
    field_subtype: Optional[str] = None
    required: bool = False

    def render(self) -> str:
        args = {}
        decl = self.name
        if self.field_subtype:
            decl += f": {self.field_type}[{self.field_subtype}]"
        decl += f" = {self.field_type}("

        if self.field_type == "ReferenceField":
            args["reference_model"] = self.field_subtype
        if self.required:
            args["required"] = "True"

        if args:
            decl += ", ".join([f"{k}={v}" for k, v in args.items()])

        decl += ")"
        return decl

    @property
    def py_type(self) -> str:
        match self.field_type:
            case "StringField":
                return "str"
            case "IntField":
                return "int"
            case "FloatField":
                return "float"
            case "BoolField":
                return "bool"
            case "ListField":
                return "List[str]"
            case "DictField":
                return "Dict[str, Any]"
            case "ReferenceField" | "SelfReferenceField" | "ObjectIdField" | "DatetimeField":
                return "str"

    def render_pydantic(self) -> str:
        return f"{self.name}: {self.py_type}"


def find_models(app_models_path: str) -> Dict[str, str]:
    from croydon.models import StorableModel

    model_imports = {}
    for _, _, files in os.walk(app_models_path):
        for filename in files:
            if filename.endswith(".py"):
                modname = "app.models." + filename[:-3].replace("/", ".")
                if modname.endswith(".__init__"):
                    modname = modname[:-9]
                module = import_module(modname)
                for item_name in dir(module):
                    item = getattr(module, item_name)
                    if inspect.isclass(item) and issubclass(item, StorableModel):
                        if item.__name__ not in model_imports:
                            model_imports[item.__name__] = item.__module__
    return model_imports


def parse_field_descriptor(fd: str) -> FieldDescriptor:
    match = FD_EXPR.findall(fd)
    if not match:
        raise ValueError(f'invalid field descriptor "{fd}"')

    tokens = match[0]

    fname, _, ftype, _, fsubtype, excl = tokens
    match ftype.lower():
        case "":
            field_type = "StringField"
        case "str" | "string" | "stringfield":
            field_type = "StringField"
        case "int" | "integer" | "integerfield":
            field_type = "IntField"
        case "bool" | "boolean" | "boolfield":
            field_type = "BoolField"
        case "float" | "floatfield":
            field_type = "FloatField"
        case "objid" | "objectid" | "objectidfield":
            field_type = "ObjectIdField"
        case "list" | "listfield":
            field_type = "ListField"
        case "dict" | "dictfield":
            field_type = "DictField"
        case "date" | "datetime" | "timestamp" | "ts" | "dt":
            field_type = "DatetimeField"
        case "ref" | "reference" | "referencefield":
            field_type = "ReferenceField"
        case "self" | "selfreference" | "selfreferencefield":
            field_type = "SelfReferenceField"
        case _:
            raise ValueError(f'unknown field type "{ftype}"')

    if fsubtype and field_type not in SUBTYPED_TYPES:
        raise ValueError(f"type {field_type} can't have subtypes")
    if not fsubtype and field_type == "ReferenceField":
        raise ValueError(f"type ReferenceField must have a subtype")

    required = False
    if excl:
        required = True

    return FieldDescriptor(
        name=fname,
        field_type=field_type,
        field_subtype=fsubtype or None,
        required=required,
    )


def render_model(
    entity_name: str,
    fields: List[str],
    models_dir: str,
    *,
    overwrite: bool = False,
    key: Optional[str] = None,
) -> None:
    filename = os.path.join(models_dir, f"{entity_name}.py")
    if not overwrite:
        if os.path.exists(filename):
            raise ValueError(
                f"file {filename} already exists, use --overwrite to overwrite it"
            )

    model_name = normalize_model_name(entity_name)
    collection = model_name.lower() + "s"
    fds = [parse_field_descriptor(field) for field in fields]

    field_types = {fd.field_type for fd in fds}
    field_subtypes = {fd.field_subtype for fd in fds if fd.field_subtype}
    subtype_imports = []
    if field_subtypes:
        models = find_models(models_dir)
        for stype in field_subtypes:
            if stype in models:
                subtype_imports.append((stype, models[stype]))

    env = Environment(loader=FileSystemLoader(get_templates_dir()))
    tmpl = env.get_template("model.py.tmpl")
    render = tmpl.render(
        field_types=field_types,
        collection=collection,
        fds=fds,
        subtype_imports=subtype_imports,
        key=key,
        class_name=model_name,
    )

    with open(filename, "w") as output:
        output.write(render)



import asyncio
import math
from time import time
from typing import Generic, List, Callable, Awaitable, Optional
from pydantic import BaseModel
from .types import TModel, T
from .db import ObjectsCursor
from . import ctx


class ItemList(BaseModel, Generic[T]):
    data: List[T]


class PaginatedList(ItemList, Generic[T]):
    count: int
    page: int
    total_pages: int


class PaginationParams(BaseModel):
    page: int
    limit: int


async def paginated(
    cur: ObjectsCursor[TModel],
    *,
    transform: Callable[[TModel], Awaitable[T]],
    params: Optional[PaginationParams] = None,
) -> ItemList[T] | PaginatedList[T]:
    t1 = time()
    if params is None:
        items = await cur.all()
        tasks = [transform(x) for x in items]
        data = await asyncio.gather(*tasks)
        ctx.log.debug(
            f"paginated {cur.ctor.__self__.__name__} list with no_paging took %.3f secs",
            time() - t1,
        )
        return ItemList(data=data)

    offset = (params.page - 1) * params.limit
    total_count = await cur.count()
    total_pages = math.ceil(total_count / params.limit)
    cur = cur.skip(offset).limit(params.limit)
    items = await cur.all()
    tasks = [transform(x) for x in items]
    data = await asyncio.gather(*tasks)
    ctx.log.debug(
        f"paginated {cur.ctor.__self__.__name__} list with limit={params.limit} took %.3f secs",
        time() - t1,
    )
    return PaginatedList(
        count=total_count, page=params.page, total_pages=total_pages, data=data
    )


async def pagination_params(
    page: int = 1, limit: Optional[int] = None, no_paging: bool = False
) -> Optional[PaginationParams]:
    if no_paging:
        return None
    if limit is None:
        limit = ctx.cfg.general.documents_per_page
    return PaginationParams(page=page, limit=limit)



from string import ascii_uppercase


def has_caps(token: str) -> bool:
    for sym in token:
        if sym in ascii_uppercase:
            return True
    return False


def normalize_model_name(token: str) -> str:
    if not has_caps(token):
        token = token.title()
    return token.replace("_", "")



import os.path
import shutil
import sys
from argparse import ArgumentParser, Namespace

import jinja2

from ..command import Command


class Init(Command):

    NAME = "init"
    HELP = "initialize a new project in the current directory"
    ASYNC_RUN = False

    _quiet: bool = False
    _overwrite: bool = False
    _project_name: str = None

    def init_argument_parser(self, parser: ArgumentParser) -> None:
        parser.add_argument("name", nargs=1, help="project name")
        parser.add_argument("-q", "--quiet", action="store_true", default=False, help="disable progress messages")
        parser.add_argument("-w", "--overwrite", action="store_true", default=False,
                            help="overwrite files and directories if they exist")

    async def setup(self, args: Namespace) -> None:
        self._quiet = args.quiet
        self._overwrite = args.overwrite
        self._project_name = args.name[0]

    def process_directory(self, dir_name: str):
        processed_files = set()

        curr_dir = os.path.abspath(os.curdir)
        for src_filename in os.listdir(dir_name):
            if src_filename == "__pycache__" or src_filename.endswith(".pyc"):
                continue

            tmpl_attrs = {"project_name": self._project_name}

            target_filename = src_filename
            full_src_filename = os.path.join(dir_name, src_filename)

            if os.path.isdir(full_src_filename):
                new_dir = os.path.join(curr_dir, target_filename)
                if new_dir in processed_files:
                    continue

                processed_files.add(new_dir)

                if not self._quiet:
                    print(f"creating new directory {new_dir}")

                create_dir = True
                if os.path.isdir(new_dir):
                    print(f"file or directory {new_dir} already exists")
                    if self._overwrite:
                        create_dir = False
                    else:
                        sys.exit(1)

                if create_dir:
                    os.makedirs(new_dir)

                if not self._quiet:
                    print(f"jumping into {new_dir}")
                os.chdir(new_dir)
                self.process_directory(full_src_filename)
                if not self._quiet:
                    print(f"jumping back into {curr_dir}")
                os.chdir(curr_dir)
            else:
                is_template = False
                new_filename = os.path.abspath(os.path.join(curr_dir, target_filename))

                if new_filename.endswith(".tmpl"):
                    new_filename = new_filename[:-5]
                    is_template = True

                if new_filename in processed_files:
                    continue

                processed_files.add(new_filename)

                if os.path.exists(new_filename) and not self._overwrite:
                    print(f"file or directory {new_filename} already exists")
                    sys.exit(1)

                if not self._quiet:
                    print(f"copying {full_src_filename} to {new_filename}")

                if is_template:
                    with open(full_src_filename) as f:
                        tmpl = jinja2.Template(f.read())
                    with open(new_filename, "w") as f:
                        f.write(tmpl.render(tmpl_attrs))
                else:
                    shutil.copy(full_src_filename, new_filename)

    def run_sync(self) -> None:

        init_command_dir = os.path.dirname(__file__)
        croydon_dir = os.path.abspath(os.path.join(init_command_dir, ".."))
        templates_dir = os.path.join(croydon_dir, "templates")
        if not self._quiet:
            print(f"using templates directory {templates_dir}")
        self.process_directory(templates_dir)
        os.chmod("crcmd.py", 0o755)



from typing import Dict, Any, TypeVar, Type, Literal, TYPE_CHECKING

if TYPE_CHECKING:
    from .config import BaseConfig
    from .models import StorableModel

T = TypeVar("T")  # any type
RT = TypeVar("RT")  # any func return type

TModel = TypeVar("TModel", bound="StorableModel")  # model type
TModelType = TypeVar("TModelType", bound=Type["StorableModel"])  # model class

QueryType = Dict[str, Any]
QueryExpression = str | QueryType

CommonDict = Dict[str, Any]
KwArgs = CommonDict

EnvironmentType = Literal["development", "testing", "staging", "production"]
CacheType = Literal["memcached", "simple", "request_local", "none"]
LogLevel = Literal["critical", "error", "warning", "info", "debug"]

TConfig = TypeVar("TConfig", bound="BaseConfig")
TConfigType = Type[TConfig]



from typing import Dict, Any, Optional
from contextvars import ContextVar
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response


REQUEST_CACHE_CONTEXT_KEY = "request_cache"

req_cache_ctx: ContextVar[Optional[Dict[str, Any]]] = ContextVar(REQUEST_CACHE_CONTEXT_KEY, default=None)


class RequestCacheMiddleware(BaseHTTPMiddleware):

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        req_cache = req_cache_ctx.set({})
        response = await call_next(request)
        req_cache_ctx.reset(req_cache)
        return response



import bcrypt
from typing import TYPE_CHECKING, Optional
from croydon import ctx
from croydon.models import StorableModel
from croydon.models.fields import StringField, DatetimeField
from croydon.util import now
if TYPE_CHECKING:
    from .token import Token


class User(StorableModel):

    COLLECTION = "users"
    KEY_FIELD = "username"

    username = StringField(required=True, unique=True)
    first_name = StringField(default="")
    last_name = StringField(default="")
    email = StringField(default="", unique=True)
    password_hash = StringField(default="-", rejected=True, restricted=True)
    avatar_url = StringField(default="")
    created_at = DatetimeField(required=True, rejected=True, default=now)
    updated_at = DatetimeField(required=True, rejected=True, default=now)

    def set_password(self, password: str) -> None:
        salt = bcrypt.gensalt()
        self.password_hash = bcrypt.hashpw(password.encode("utf-8"), salt).decode("utf-8")

    def check_password(self, password: str) -> bool:
        try:
            return bcrypt.checkpw(
                password.encode("utf-8"),
                self.password_hash.encode("utf-8")
            )
        except ValueError as e:
            ctx.log.error(e)
            # password not set leads to bcrypt raising ValueError("invalid salt")
            return False

    async def create_token(self, *, description: Optional[str] = None) -> "Token":
        from .token import Token
        t = Token({"user_id": self.id, "description": description})
        await t.save()
        return t



import os.path
from argparse import ArgumentParser, Namespace
from croydon.command import Command
from croydon.project.util import normalize_model_name
from croydon.project.models import parse_field_descriptor, find_models
from croydon import ctx


class New(Command):
    NAME = "new"
    HELP = "create new entities"
    ASYNC_RUN = False

    args: Namespace
    entity_name: str

    def init_argument_parser(self, parser: ArgumentParser) -> None:
        sub = parser.add_subparsers(dest="entity", required=True)
        model = sub.add_parser(name="model")
        model.add_argument("name", nargs=1, help="model name")
        model.add_argument("fields", nargs="*", help="model fields configuration")
        model.add_argument("-k", "--key", type=str, help="key field name")
        model.add_argument(
            "-w",
            "--overwrite",
            action="store_true",
            help="overwrite existing entity if it exists",
        )

        controller = sub.add_parser(name="controller")
        controller.add_argument("name", nargs=1, help="controller name")
        controller.add_argument("-p", "--prefix", help="api prefix", default="/api/v1")
        controller.add_argument(
            "-w",
            "--overwrite",
            action="store_true",
            help="overwrite existing entity if it exists",
        )

    async def setup(self, args: Namespace) -> None:
        self.args = args
        self.entity_name = self.args.name[0]

    def setup_model(self) -> None:
        models_dir = os.path.join(ctx.project_dir, "app/models")
        filename = os.path.join(models_dir, f"{self.entity_name}.py")
        if not self.args.overwrite:
            if os.path.exists(filename):
                raise ValueError(
                    f"file {filename} already exists, use --overwrite to overwrite it"
                )

        model_name = normalize_model_name(self.entity_name)
        collection = model_name.lower() + "s"
        fds = [parse_field_descriptor(field) for field in self.args.fields]

        field_types = {fd.field_type for fd in fds}
        field_subtypes = {fd.field_subtype for fd in fds if fd.field_subtype}

        with open(filename, "w") as output:
            output.write("from croydon.models import StorableModel\n")
            output.write(
                "from croydon.models.fields import " + ", ".join(field_types) + "\n"
            )

            if field_subtypes:
                models = find_models(models_dir)
                for stype in field_subtypes:
                    if stype in models:
                        output.write(f"from {models[stype]} import {stype}\n")

            output.write("\n\n")
            output.write(f"class {model_name}(StorableModel):\n\n")
            output.write(f'    COLLECTION = "{collection}"\n')
            if self.args.key:
                output.write(f'    KEY_FIELD = "{self.args.key}"\n\n')
            for fd in fds:
                output.write(f"    {fd.render()}\n")

    def setup_controller(self) -> None:
        raise NotImplementedError("controller setup is not implemented yet")

    def run_sync(self) -> None:
        try:
            match self.args.entity:
                case "model":
                    self.setup_model()
                case "controller":
                    self.setup_controller()
        except ValueError as e:
            print(e)



from typing import Literal, Optional
from datetime import datetime
from pydantic import BaseModel
from . import ObjectIdStr


class AccountMeResponse(BaseModel):
    id: Optional[ObjectIdStr]
    username: Optional[str]
    first_name: Optional[str]
    last_name: Optional[str]
    email: Optional[str]
    ext_id: Optional[str]
    avatar_url: Optional[str]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]


class LogoutResponse(BaseModel):
    detail: Literal["logged out"] = "logged out"


class AuthenticationRequest(BaseModel):
    username: str
    password: str
